/*
Copyright 1990-2001 Sun Microsystems, Inc. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions: The above copyright notice and this
permission notice shall be included in all copies or substantial
portions of the Software.


THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE OPEN GROUP OR SUN MICROSYSTEMS, INC. BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE EVEN IF
ADVISED IN ADVANCE OF THE POSSIBILITY OF SUCH DAMAGES.


Except as contained in this notice, the names of The Open Group and/or
Sun Microsystems, Inc. shall not be used in advertising or otherwise to
promote the sale, use or other dealings in this Software without prior
written authorization from The Open Group and/or Sun Microsystems,
Inc., as applicable.


X Window System is a trademark of The Open Group

OSF/1, OSF/Motif and Motif are registered trademarks, and OSF, the OSF
logo, LBX, X Window System, and Xinerama are trademarks of the Open
Group. All other trademarks and registered trademarks mentioned herein
are the property of their respective owners. No right, title or
interest in or to any trademark, service mark, logo or trade name of
Sun Microsystems, Inc. or its licensors is granted.

*/

#include "iiimpReq.h"
#include "iiimpIM.h"
#include "guiIM.h"
#include "XimpIm.h"
#include "xiiimp.h"
#include "iiimpAuth.h"
#include "iiimpAuxP.h"
#include "iiimpUtil.h"
#include "iiimpColor.h"
#include "iiimpTr.h"
#include "iiimpConv.h"
#include "iiimpSwitcher.h"
#include "KeyMap.h"
#include "trace_message.h"
#include <stdio.h>
#include <locale.h>
#include <string.h>
#include <sys/utsname.h>
#include <iiimcf.h>
#include <langinfo.h>
#include <limits.h>
#include <values.h>
#include <iconv.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

static IIIMCF_language get_input_language (XicCommon ic, char *input_language,
					   Bool exact_match);

static IIIMF_status
process_keyevent(
    XicCommon ic,
    IIIMCF_event ev
)
{
    IIIMF_status st;
    IIIMCF_keyevent kev;
    XKeyEvent xkev;
    XimCommon im = (XimCommon)ic->core.im;

    unsigned int keysym;
    unsigned short ucode;

    Bool useUnicode = (XIM_USE_UNICODE(im) &&
		  ic->commit_string_callback.callback);

    st = iiimcf_get_keyevent_value(ev, &kev);
    if (st != IIIMF_STATUS_SUCCESS) return st;

    ucode = kev.keychar;
    keysym = unicode_to_keysym(ucode);

    /* this keysym is invalid, try to get the keysym by iiim keycode */
    if (keysym & 0x01000000) {
        keysym = i2gcode (kev.keycode);
        ucode = keysym_to_unicode (keysym);
    }

    if (kev.modifier & 0x40000000 &&
	keysym >= 0x61 && keysym <= 0x7a) {
        keysym -= 0x20;
    }

    /* if the key is not a release event, nor combined key, such as ctrl+u,
     * and the key represents a single unicode character, try to commit it.
     */
    if ( 0 == (kev.modifier & 0x8000000e) && ucode)
      {
        /* for Java applications */
        if (ic->commit_string_callback.callback)
          {
            XIMCallback *cb = &ic->commit_string_callback;
            XIMText cbtext;
            char *str;

            UTF16ToString (1, &ucode, useUnicode, &str);

            memset (&cbtext, 0, sizeof (XIMText));
            cbtext.length = 1; /* not the byte length */
            cbtext.string.multi_byte = str;

            (*cb->callback)((XIC)ic,
                            cb->client_data,
                            (XPointer)&cbtext);

            /* no need to fake a keyevent to be putback */
            if (cbtext.string.multi_byte) Xfree(cbtext.string.multi_byte);
            if (cbtext.feedback) Xfree(cbtext.feedback);
      
            return IIIMF_STATUS_SUCCESS;
          } 
        else
          {
            XIC_IIIMP(ic, keysym_return) = keysym;
            XIC_IIIMP(ic, ucode_return) = ucode;
            XIC_IIIMP(ic, committed_text_consumed) = False;

            /* fake a key press event, which keycode is 0, so that the keyfilter 
             * would return false, which in turn causes xim client to call lookup
             * string function. The code section in lookup_string will try to
             * convert this keysym to utf8 character, and copy to the client's
             * buffer.
             */
            memset(&xkev, 0, sizeof(xkev));
            xkev.type = (int)KeyPress;
            xkev.display = im->core.display;
            xkev.window = ic->core.focus_window;
            xkev.root = None;
            xkev.subwindow = None;
            xkev.keycode = XIM_COMPOSE_KEYCODE;
            xkev.same_screen = True;
            /* XPutBackEvent(im->core.display, &ev); */
            /* ic->ximp_icpart->putback_key_event = True; */
            HoldXKeyEvent(ic, (XEvent*)&xkev);

            return IIIMF_STATUS_SUCCESS;
          }
      }

    IIIMCF_keyevent_To_XKeyEvent(&kev, im->core.display, ic->core.focus_window, &xkev);
    
    if (ic->forward_event_callback.callback) {
	ic->forward_event_callback.callback((XIC)ic,
					    ic->forward_event_callback.client_data,
					    (XPointer)&xkev);
    } else {
	HoldXKeyEvent(ic, (XEvent*)&xkev);
    }

    return IIIMF_STATUS_SUCCESS;
}

static IIIMF_status
process_trigger_notify(
    XicCommon ic,
    IIIMCF_event ev
)
{
    IIIMF_status st;
    int flag;

    st = iiimcf_get_trigger_notify_flag(ev, &flag);
    if (st != IIIMF_STATUS_SUCCESS) return st;
    if (flag)
	SetConversionMode(ic, True);
    else
	SetConversionMode(ic, False);

    return IIIMF_STATUS_SUCCESS;
}
    
static IIIMF_status
draw_preedit(
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_text text;
    XIMText cbtext;
    int caret = 0, len;
    IMFeedbackList *feedback_list = NULL;
    XICCallback *cb = &ic->core.preedit_attr.draw_callback;
    XIMPreeditDrawCallbackStruct call_data;
    XimCommon im = (XimCommon)ic->core.im;
    Bool useUnicode;

    memset(&call_data, 0, sizeof(XIMPreeditDrawCallbackStruct));
    memset(&cbtext, 0, sizeof(XIMText));

    useUnicode = (XIM_USE_UNICODE(im) && cb->callback &&
		  (ic->core.input_style & XIMPreeditCallbacks));

    st = iiimcf_get_preedit_text(XIC_IIIMP(ic, context), &text, &caret);
    if (st == IIIMF_STATUS_NO_PREEDIT) {
	call_data.text = NULL;
	len = 0;
    } else {
	if (st != IIIMF_STATUS_SUCCESS) return st;
	st = iiimcf_get_text_length(text, &len);
	if (st != IIIMF_STATUS_SUCCESS) return st;
	if (len > 0) {
	    st = IIimpConvertToXIMText(text, &cbtext, &feedback_list, useUnicode);
	    if (st != IIIMF_STATUS_SUCCESS) return st;
	    call_data.text = &cbtext;
	} else {
	    call_data.text = NULL;
	}
    }
    call_data.caret = caret;
    call_data.chg_first = 0;
    call_data.chg_length = XIC_IIIMP(ic, preedit_prev_len);
    CallPreeditDrawCallback(ic, &call_data, feedback_list);
    FreeFeedbackList(feedback_list, cbtext.length);
    XIC_IIIMP(ic, preedit_prev_len) = len;
    if (cbtext.string.multi_byte) Xfree(cbtext.string.multi_byte);
    if (cbtext.feedback) Xfree(cbtext.feedback);

    return IIIMF_STATUS_SUCCESS;
}

static IIIMF_status
draw_status(
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_text text;
    XIMText cbtext;
    IMFeedbackList *feedback_list = NULL;
    XICCallback *cb = &ic->core.status_attr.draw_callback;
    XIMStatusDrawCallbackStruct call_data;
    XimCommon im = (XimCommon)ic->core.im;
    Bool useUnicode;

    memset(&call_data, 0, sizeof(XIMStatusDrawCallbackStruct));
    memset(&cbtext, 0, sizeof(XIMText));

    useUnicode = (XIM_USE_UNICODE(im) && cb->callback &&
		  (ic->core.input_style & XIMStatusCallbacks));

    st = iiimcf_get_status_text(XIC_IIIMP(ic, context), &text);
    if (st == IIIMF_STATUS_NO_STATUS_TEXT) {
	/* erase status text.  */
	cbtext.feedback = NULL;
	cbtext.string.multi_byte = NULL;
	cbtext.length = 0;
	call_data.data.text = &cbtext;
    } else {
	if (st != IIIMF_STATUS_SUCCESS) return st;
	st = IIimpConvertToXIMText(text, &cbtext, &feedback_list, useUnicode);
	if (st != IIIMF_STATUS_SUCCESS) return st;
	call_data.data.text = &cbtext;
	call_data.type = XIMTextType;
    }
    CallStatusDrawCallback(ic, &call_data, feedback_list, useUnicode);
    FreeFeedbackList(feedback_list, cbtext.length);
    if (cbtext.string.multi_byte) Xfree(cbtext.string.multi_byte);
    if (cbtext.feedback) Xfree(cbtext.feedback);

    return IIIMF_STATUS_SUCCESS;
}

static IIIMF_status
start_lookup_choice(
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_lookup_choice lc;
    XIMLookupStartCallbackStruct2 call_data;
    int direction;

    st = iiimcf_get_lookup_choice(XIC_IIIMP(ic, context), &lc);
    if (st != IIIMF_STATUS_SUCCESS) return st;
    st = iiimcf_get_lookup_choice_configuration(lc,
						&call_data.choice_per_window,
						&call_data.nrows,
						&call_data.ncolumns,
						&direction);
    if (direction == IIIMCF_LOOKUP_CHOICE_HORIZONTAL_DIRECTION) {
	call_data.draw_up_direction = XIMDrawUpHorizontally;
    } else {
	call_data.draw_up_direction = XIMDrawUpVertically;
    }
    if (st != IIIMF_STATUS_SUCCESS) return st;
    /* libiiimcf rules them.  */
    call_data.which_is_master = IM_LOOKUP_CHOICE_START_SERVER_IS_MASTER;
    call_data.who_owns_label = 0;

    if (ic->lookup_start_callback.callback) {
	XIMCallback *cb = &ic->lookup_start_callback;
	(*cb->callback)((XIC)ic, cb->client_data, (XPointer)&call_data);
    } else {
	XIC_GUI(ic, change_lookup)((XIC)ic, LOOKUP_START,
				   (XPointer)&call_data);
    }

    return IIIMF_STATUS_SUCCESS;
}

static IIIMF_status
draw_lookup_choice(
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_lookup_choice lc;
    IIIMCF_text title, label, cand;
    XIMLookupDrawCallbackStruct2 call_data;
    XIMChoiceObject2 choice_object;
    XIMText *choice_list = (XIMText*)NULL, *label_list = (XIMText*)NULL;
    XIMText title_text;
    IMFeedbackList **choice_feedback = (IMFeedbackList**)NULL;
    IMFeedbackList **label_feedback = (IMFeedbackList**)NULL;
    XimCommon im = (XimCommon)ic->core.im;
    Bool useUnicode;
    int size, flag;
    int i;

    st = iiimcf_get_lookup_choice(XIC_IIIMP(ic, context), &lc);
    if (st != IIIMF_STATUS_SUCCESS) return st;
    st = iiimcf_get_lookup_choice_size(lc, &size,
				       &call_data.first_index,
				       &call_data.last_index,
				       &call_data.current_index);
    if (st != IIIMF_STATUS_SUCCESS) return st;
    if (size <= 0) return IIIMF_STATUS_SUCCESS;
    st = iiimcf_get_lookup_choice_title(lc, &title);
    if (st != IIIMF_STATUS_SUCCESS) return st;

    useUnicode = (XIM_USE_UNICODE(im) &&
		  ic->lookup_draw_callback.callback);

    choice_list = (XIMText*) Xmalloc(sizeof(XIMText) * size * 2);
    if (!choice_list) {
	st = IIIMF_STATUS_MALLOC;
	goto cleanup;
    }
    memset(choice_list, 0, sizeof(XIMText) * size * 2);
    label_list = choice_list + size;

    choice_feedback = (IMFeedbackList**) Xmalloc(sizeof(IMFeedbackList*) * size * 2);
    if (!choice_feedback) {
	st = IIIMF_STATUS_MALLOC;
	goto cleanup;
    }
    memset(choice_feedback, 0, sizeof(IMFeedbackList*) * size * 2);
    label_feedback = choice_feedback + size;

    memset(&title_text, 0, sizeof(XIMText));
    st = IIimpConvertToXIMText(title, &title_text, NULL, useUnicode);
    if (st != IIIMF_STATUS_SUCCESS) goto cleanup;

    for (i = 0; i < size; i++) {
	st = iiimcf_get_lookup_choice_item(lc, i, &cand, &label, &flag);
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
	st = IIimpConvertToXIMText(cand, &choice_list[i],
				   &choice_feedback[i], useUnicode);
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
	st = IIimpConvertToXIMText(label, &label_list[i],
				   &label_feedback[i], useUnicode);
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
    }

    call_data.n_choices = size;
    choice_object.label_decoration = NO_DECORATION;
    choice_object.value = choice_list;
    choice_object.label = label_list;
    choice_object.value_feedback = choice_feedback;
    choice_object.label_feedback = label_feedback;
    call_data.choices = &choice_object;
    call_data.title = &title_text;

    if (ic->lookup_draw_callback.callback) {
	XIMCallback *cb = &ic->lookup_draw_callback;
	(*cb->callback)((XIC)ic, cb->client_data, (XPointer)&call_data);
    } else {
	XIC_GUI(ic, change_lookup)((XIC)ic, LOOKUP_DRAW,
				   (XPointer)&call_data);
    }

cleanup:

    for (i = 0; i < size; i++) {
	if (choice_list) {
	    if (choice_feedback) {
		FreeFeedbackList(choice_feedback[i], choice_list[i].length);
	    }
	    if (choice_list[i].string.multi_byte) {
		Xfree(choice_list[i].string.multi_byte);
	    }
	    if (choice_list[i].feedback) {
		Xfree(choice_list[i].feedback);
	    }
	}
	if (label_list) {
	    if (label_feedback) {
		FreeFeedbackList(label_feedback[i], label_list[i].length);
	    }
	    if (label_list[i].string.multi_byte) {
		Xfree(label_list[i].string.multi_byte);
	    }
	    if (label_list[i].feedback) {
		Xfree(label_list[i].feedback);
	    }
	}
    }
    if (choice_list) Xfree(choice_list);
    if (choice_feedback) Xfree(choice_feedback);

    if (title_text.string.multi_byte) {
	Xfree(title_text.string.multi_byte);
    }
    if (title_text.feedback) {
	Xfree(title_text.feedback);
    }

    return st;
}

static IIIMF_status
end_lookup_choice(
    XicCommon ic
)
{
    if (ic->lookup_done_callback.callback) {
	XIMCallback *cb = &ic->lookup_done_callback;
	(*cb->callback)((XIC)ic, cb->client_data, (XPointer)NULL);
    } else {
	XIC_GUI(ic, change_lookup)((XIC)ic, LOOKUP_DONE, (XPointer)NULL);
    }

    return IIIMF_STATUS_SUCCESS;
}

static IIIMF_status
commit(
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_text text;
    XEvent ev;
    Bool useUnicode;
    XimCommon im = (XimCommon)ic->core.im;

    useUnicode = (XIM_USE_UNICODE(im) &&
		  ic->commit_string_callback.callback);

    /* private XIC extension */
    if (ic->commit_string_callback.callback) {
	XIMText cbtext;
	XIMCallback *cb = &ic->commit_string_callback;

	st = iiimcf_get_committed_text(XIC_IIIMP(ic, context), &text);
	if (st != IIIMF_STATUS_SUCCESS) return st;
	st = IIimpConvertToXIMText(text, &cbtext, NULL, useUnicode);
	if (st != IIIMF_STATUS_SUCCESS) return st;
	(*cb->callback)((XIC)ic,
			cb->client_data,
			(XPointer)&cbtext);	/* for UNICODE1 */
	/* no need to fake a keyevent to be putback */
	if (cbtext.string.multi_byte) Xfree(cbtext.string.multi_byte);
	if (cbtext.feedback) Xfree(cbtext.feedback);
	return IIIMF_STATUS_SUCCESS;
    }

    /* fake a key press event */
    memset(&ev, 0, sizeof(ev));
    ev.type = (int)KeyPress;
    ev.xkey.display = im->core.display;
    ev.xkey.window = ic->core.focus_window;
    ev.xkey.root = None;
    ev.xkey.subwindow = None;
    ev.xkey.keycode = XIM_COMPOSE_KEYCODE;
    ev.xkey.same_screen = True;
    /* XPutBackEvent(im->core.display, &ev); */
    /* ic->ximp_icpart->putback_key_event = True; */
    HoldXKeyEvent(ic, &ev);
    XIC_IIIMP(ic, keysym_return) = NoSymbol;
    XIC_IIIMP(ic, ucode_return) = 0;
    XIC_IIIMP(ic, committed_text_consumed) = False;

    return IIIMF_STATUS_SUCCESS;
}

IIIMF_status
IMProcessIncomingEvent(
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_event ev;
    IIIMCF_event_type et;
    IIIMCF_context c;

    if (!ic) return IIIMF_STATUS_SUCCESS;

    c = XIC_IIIMP(ic, context);
    while ((st = iiimcf_get_next_event(c, &ev))
	   == IIIMF_STATUS_SUCCESS) {
	st = iiimcf_dispatch_event(c, ev);
	if (st != IIIMF_STATUS_SUCCESS) {
	    if (iiimcf_get_event_type(ev, &et) != IIIMF_STATUS_SUCCESS)
		continue;
	    switch (et) {
	      case IIIMCF_EVENT_TYPE_KEYEVENT:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_KEYEVENT\n"));
	       process_keyevent(ic, ev);
	       break;

	      case IIIMCF_EVENT_TYPE_TRIGGER_NOTIFY:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_TRIGGER_NOTIFY\n"));
	       process_trigger_notify(ic, ev);
	       ic_switcher_set_conversion_mode (ic);
	       break;

	      case IIIMCF_EVENT_TYPE_UI_PREEDIT_START:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_PREEDIT_START\n"));
	       CallPreeditStartCallback(ic, NULL);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_PREEDIT_CHANGE:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_PREEDIT_CHANGE\n"));
	       draw_preedit(ic);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_PREEDIT_DONE:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_PREEDIT_DONE\n"));
	       CallPreeditDoneCallback(ic, NULL);
	       break;

	      case IIIMCF_EVENT_TYPE_UI_STATUS_START:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_STATUS_START\n"));
	       CallStatusStartCallback(ic, NULL);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_STATUS_CHANGE:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_STATUS_CHANGE\n"));
	       draw_status(ic);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_STATUS_DONE:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_STATUS_DONE\n"));
	       CallStatusDoneCallback(ic, NULL);
	       break;

	      case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_START:
	       start_lookup_choice(ic);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_CHANGE:
	       draw_lookup_choice(ic);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_DONE:
	       end_lookup_choice(ic);
	       break;

	      case IIIMCF_EVENT_TYPE_AUX_START:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_AUX_START\n"));
	       IIimpAuxStart(ic, ev);
	       break;
	      case IIIMCF_EVENT_TYPE_AUX_DRAW:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_AUX_DRAW\n"));
	       IIimpAuxDraw(ic, ev);
	       break;
	      case IIIMCF_EVENT_TYPE_AUX_DONE:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_AUX_DONE\n"));
	       IIimpAuxDone(ic, ev);
	       break;

	      case IIIMCF_EVENT_TYPE_AUX_GETVALUES:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_AUX_GETVALUES\n"));
	       IIimpAuxGetValuesReply(ic, ev);
	       break;
	      case IIIMCF_EVENT_TYPE_UI_COMMIT:
	       TRACE_MESSAGE('i', ("IIIMCF_EVENT_TYPE_UI_COMMIT\n"));
	       commit(ic);
	       break;
	      default:
	       TRACE_MESSAGE('i', ("unhandled IIIMCF_EVENT_TYPE\n"));
	       break;
	    }
	    /* discard event */
	    iiimcf_ignore_event(ev);
	}
    }

    PutBackXKeyEvent(ic);

    return IIIMF_STATUS_SUCCESS;
}

Bool
IMTriggerNotify(
    XicCommon ic,
    XICConversionMode conv_mode
)
{
    IIIMCF_event ev;
    if (!ic) return False;

    if (conv_mode == CONV_OFF) {
      end_lookup_choice(ic);
    }

    if (iiimcf_create_trigger_notify_event(conv_mode, &ev)
	!= IIIMF_STATUS_SUCCESS) return False;

    if (iiimcf_forward_event(XIC_IIIMP(ic, context), ev)
	!= IIIMF_STATUS_SUCCESS) return False;
    IMProcessIncomingEvent(ic);

    return True;
}

Bool
IMForwardEvent(
    XicCommon ic,
    XEvent *ev
)
{
    IIIMF_status st;
    IIIMCF_keyevent kev;
    IIIMCF_event event;

    if (!XKeyEvent_To_IIIMCF_keyevent((XKeyEvent*) ev, &kev))
	return True;

#if 0 /* SHOULD BE REMOVED */
    fprintf(stderr, "(%d, %d)->(%d, %d, %d)\n",
	    ev->xkey.keycode, ev->xkey.state,
	    kev.keycode, kev.keychar, kev.modifier);
#endif

    st = iiimcf_create_keyevent(&kev, &event);
    if (st != IIIMF_STATUS_SUCCESS) return False;

    st = iiimcf_forward_event(XIC_IIIMP(ic, context), event);

    IMProcessIncomingEvent(ic);

    if (st != IIIMF_STATUS_SUCCESS) return False;

    return True;
}

Atom vkb_target_atom;

static void
SetVKBTarget(XicCommon ic) {
    RegisterVKBFilter(ic);
    XSetSelectionOwner(ic->core.im->core.display, vkb_target_atom, ic->core.client_window, CurrentTime);
}

static void
UnsetVKBTarget(XicCommon ic) {
  UnregisterVKBFilter(ic);
}

IIIMF_status
IMChangeFocus(
    XicCommon ic,
    int flag
)
{
    IIIMF_status st;
    IIIMCF_event event;

    if (ic == NULL || ic->iiimp_icpart == NULL)
        return IIIMF_STATUS_SUCCESS;

    if (flag) {
	st = iiimcf_create_seticfocus_event(&event);
	IIimpAuxSetICFocus(ic);
    } else {
	st = iiimcf_create_unseticfocus_event(&event);
	IIimpAuxUnsetICFocus(ic);
    }
    if (st != IIIMF_STATUS_SUCCESS) return st;

    if (!flag) {
      SetVKBTarget(ic);
    } else {
      UnsetVKBTarget(ic);
    }

    st = iiimcf_forward_event(XIC_IIIMP(ic, context), event);
    IMProcessIncomingEvent(ic);

    /* to reflect current lang_id to gimlet */
    if (flag) {
      ic_switcher_set_input_language (ic, NULL);
    }

    return st;
}

#ifdef HAS_IIIM_PROPERTIES
void
IM_free_langlist(
    XimCommon im
)
{
    int i;
    int n = XIM_IIIMP(im, count_languages);
    LangElement *language_list = XIM_IIIMP(im, supported_languages);
    char *str;

    if (language_list) {
        for (i = 0; i < n; i++) {
	    Xfree (language_list[i].lang_id);
	    Xfree (language_list[i].lang_ui);
	    Xfree (language_list[i].engine_id);
	}
	Xfree (language_list);
    }
    XIM_IIIMP(im, count_languages) = 0;
    XIM_IIIMP(im, supported_languages) = NULL;
}
#else /* HAS_IIIM_PROPERTIES */
void
IM_free_langlist(
    XimCommon im
)
{
    int i;
    int n = XIM_IIIMP(im, count_languages);
    XIMText *language_list = XIM_IIIMP(im, supported_languages);
    char *str;

    if (language_list) {
	for (i = 0; i < n; i++) {
	    str = language_list[i].string.multi_byte;
	    if (str) free(str);
	}
	free(language_list);
    }
    XIM_IIIMP(im, count_languages) = 0;
    XIM_IIIMP(im, supported_languages) = NULL;
}
#endif /* HAS_IIIM_PROPERTIES */

#ifdef HAS_IIIM_PROPERTIES

static char *
fetch_element (char *p) {
    char *ptr, *ret;
    int len = 0;
    for (ptr = p; *ptr; ptr++) {
        if (*ptr == ';') {
	    len = ptr - p;
	    ret = Xmalloc (sizeof (char) * len + 1);
	    strncpy (ret, p, len);
	    ret[len] = '\0';
	    break;
	}
    }
    return ret;
}

static void
make_langlist_with_alloc (unsigned char *list, int *count, LangElement **lang_list)
{
    int i;
    int delim_count = 0;
    char *p, *part;
    int list_len = strlen ((const char *)list);
    
    for (i = 0; i < list_len; i++) {
        if (list[i] == ';') {
	   delim_count++;
	}
    }
    /* list is separated three ";" */
    *count = delim_count / 3; 
    
    *lang_list = (LangElement *)Xmalloc(sizeof(LangElement) * (*count));

    p = (char *)list;
    for (i = 0; i < *count; i++) {
        part = fetch_element(p);
        (*lang_list)[i].lang_id = part;
	p += strlen (part) + 1;
	part = fetch_element(p);
	(*lang_list)[i].lang_ui = part;
	p += strlen (part) + 1;
	part = fetch_element (p);
	(*lang_list)[i].engine_id = part;
	p += strlen (part) + 1;
    }
}

#endif /* HAS_IIIM_PROPERTIES */

static void
init_langlist(
    XimCommon im
)
{
#ifdef HAS_IIIM_PROPERTIES
    /* try to get language list from user's preference which is made
       with iiim-properties tool */
    Atom lang_list_atom;
    XTextProperty text_props;
    LangElement *language_list;
    int n;

    text_props.value = NULL;
    lang_list_atom = XInternAtom (im->core.display, "_IIIM_SWITCHER_INPUT_LANGUAGE_LIST", True);
    if (lang_list_atom != None) {
       XGetTextProperty (im->core.display, 
			 RootWindow (im->core.display, DefaultScreen (im->core.display)),
			 &text_props, lang_list_atom);
    }
    if (text_props.value != NULL) {
       make_langlist_with_alloc (text_props.value, &n, &language_list);
       IM_free_langlist(im);
       XIM_IIIMP(im, count_languages) = n;
       XIM_IIIMP(im, supported_languages) = language_list;
       /* do not use below code which gets all of supported langs from server */
       free(text_props.value);
       return;
    }
#else /* HAS_IIIM_PROPERTIES */
    int i, n, len;
    const char *ll;
    char *ll2;
    IIIMCF_language *plangs;
    XIMText *language_list = NULL;

    IM_free_langlist(im);
    if (iiimcf_get_supported_languages(XIM_IIIMP(im, handle), &n, &plangs)
	!= IIIMF_STATUS_SUCCESS)
	return;
    language_list = (XIMText*) Xmalloc(sizeof(XIMText) * n);
    if (!language_list) return;
    memset(language_list, 0, sizeof(XIMText) * n);
    XIM_IIIMP(im, count_languages) = n;
    XIM_IIIMP(im, supported_languages) = language_list;
    for (i = 0; i < n; i++) {
	iiimcf_get_language_id(plangs[i], &ll);
	len = strlen(ll);
	ll2 = (char*) Xmalloc(sizeof(char) * (len + 1));
	if (!ll2) {
	    IM_free_langlist(im);
	    return;
	}
	memcpy(ll2, ll, sizeof(char) * (len + 1));
	language_list[i].length = len;
	language_list[i].string.multi_byte = ll2;
    }
#endif /* HAS_IIIM_PROPERTIES */
    return;
}

static IIIMF_status
IMSetupAuxObject(
    XimCommon im
);

#ifdef HAS_IIIM_PROPERTIES
Atom global_conv_mode_atom;
#endif /* HAS_IIIM_PROPERTIES */
IIIMF_status
IMCreateHandle(
    XimCommon im
)
{
    IIIMF_status st;
    IIIMCF_attr attr;
    char *address, *service;
    unsigned char * client_group_id;
    Atom iiimd;
    Window iiimx;
    Atom client_group;
    Atom type;
    int format;
    unsigned long length;
    unsigned long nitem;
    unsigned char *prop;

    address = NULL;
    service = NULL;
    prop = NULL;
    client_group_id = NULL;

    st = iiimcf_create_attr(&attr);
    if (st != IIIMF_STATUS_SUCCESS) return st;
    if (!XIM_IIIMP(im, client_type)) {
	XIM_IIIMP(im, client_type) = IIIMP_CLIENT_TYPE; /* default */
    }
    st = iiimcf_attr_put_string_value(attr, IIIMCF_ATTR_CLIENT_TYPE,
				      XIM_IIIMP(im, client_type));
    if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
    if (im->core.display) {
	Display *dpy = im->core.display;

	st = iiimcf_attr_put_string_value(attr, IIIMCF_ATTR_CLIENT_X_DISPLAY_NAME,
					  XDisplayString(dpy));
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
	st = iiimcf_attr_put_string_value(attr, IIIMCF_ATTR_CLIENT_X_SERVER_VENDOR,
					  XServerVendor(dpy));
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
    }

    if (NULL != im->core.display) {
	iiimd = XInternAtom (im->core.display, "IIIM_SERVER", True);
	if (None != iiimd) {
	    iiimx = XGetSelectionOwner (im->core.display, iiimd);
	} else {
	    iiimx = None;
	}
	if (None != iiimx) {
	    XGetWindowProperty(im->core.display, iiimx, iiimd,
			       0, INT_MAX, False, XA_STRING,
			       &type, &format, &nitem, &length, &prop);
	    if (NULL == prop) {
	    } else if (0 == strncmp("uds:", (char *)prop, 4)) {
		address = strdup((prop + 4));
		service = "";
		Xfree(prop);
		prop = NULL;
		if (NULL == address) goto cleanup;
	    } else {
		Xfree(prop);
		prop = NULL;
	    }
	}

	client_group = XInternAtom(im->core.display, "IIIM_CLIENT_GROUP", True);
	if (None != client_group) {
	    iiimx = XGetSelectionOwner(im->core.display, client_group);
	} else {
	    iiimx = None;
	}
	if (None != iiimx) {
	    XGetWindowProperty(im->core.display, iiimx, client_group,
			       0, INT_MAX, False, XA_STRING,
			       &type, &format, &nitem, &length, &prop);
	    if (NULL != prop) {
		client_group_id = strdup(prop);
		Xfree(prop);
		prop = NULL;
	    }
	}
    }
    if ((NULL == address) && (im->core.im_name)) {
	address = strdup(im->core.im_name);
	if (!address) goto cleanup;
	service = index(address, ':');
	if (service) {
	    *service = '\0';
	    service++;
	}
    }

    if (NULL != address) {
	st = iiimcf_attr_put_string_value(attr, IIIMCF_ATTR_SERVER_ADDRESS,
					  address);
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
	if (service) {
	    st = iiimcf_attr_put_string_value(attr, IIIMCF_ATTR_SERVER_SERVICE,
					      service);
	    if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
	}
    }

    if (NULL != client_group_id) {
	st = iiimcf_attr_put_string_value(attr, IIIMCF_ATTR_CLIENT_GROUP,
					  client_group_id);
	if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
    }

    st = iiimcf_create_handle(attr, &XIM_IIIMP(im, handle));
    if (st != IIIMF_STATUS_SUCCESS) goto cleanup;
    init_langlist(im);
    RegisterTriggerKeys(im);
    IMSetupAuxObject(im);

#ifdef HAS_IIIM_PROPERTIES
    global_conv_mode_atom = XInternAtom (im->core.display, "_IIIM_GLOBAL_CONV_MODE", False);
#endif /* HAS_IIIM_PROPERTIES */    
    vkb_target_atom = XInternAtom(im->core.display, "_IIIM_VKB_TARGET", False);

cleanup:
    iiimcf_destroy_attr(attr);
    if (address) free(address);
    if (client_group_id) free(client_group_id);
    if (NULL != prop) XFree(prop);
    return st;
}

static Bool
vkb_commit_filter(Display *display, Window window, XEvent *eventp, XPointer context) {
    IIIMF_status st;
    XicCommon ic;
    IIIMCF_context ctx;

    if (eventp->type != ClientMessage || eventp->xclient.message_type != vkb_target_atom) {
        /* This is not event from VKB */
        return False;
    }

    ic = (XicCommon)context;
    ctx = XIC_IIIMP(ic, context);

    /* get stored commit string here */
    iiimcf_forward_event_with_operations (ctx, NULL, IIIMCF_GET_STORED_CONTENTS);
    IMProcessIncomingEvent(ic);
  
    return True;
}

IIIMF_status
IMCreateIC(
    XimCommon im,
    XicCommon ic
)
{
    IIIMF_status st;
    IIIMCF_attr attr;
    IIIMCF_language lang;
    
    st = iiimcf_create_attr(&attr); 
    if (st != IIIMF_STATUS_SUCCESS) return st;
    lang = get_IM_language(im);
    if (lang) {
	st = iiimcf_attr_put_ptr_value(attr,
				       IIIMCF_ATTR_INPUT_LANGUAGE,
				       lang);
	if (st != IIIMF_STATUS_SUCCESS) goto exit;
    }

    if (XIM_IIIMP(im, engine_name)) {
	st = iiimcf_attr_put_string_value(attr,
					  IIIMCF_ATTR_INPUT_METHOD_NAME,
					  XIM_IIIMP(im, engine_name));
	if (st != IIIMF_STATUS_SUCCESS) goto exit;
    }
#ifdef HAS_IIIM_PROPERTIES
    st = iiimcf_attr_put_integer_value(attr, IIIMCF_ATTR_KBD_LAYOUT, GetKBDLayout ());
    if (st != IIIMF_STATUS_SUCCESS) {
      goto exit;
    }
    /* Default LE handling */
    {
      char *lang_id, *engine_id = NULL;
      extern char *im_canonicalize_langname(XimCommon);
      if (lang == NULL) {
	lang_id = im_canonicalize_langname (im);
      } else {
	st = iiimcf_get_language_id (lang, (const char **)&lang_id);
	if (st != IIIMF_STATUS_SUCCESS) {
	  lang_id = im_canonicalize_langname (im);
	}
      }

      Xfree (ic->current_language);
      ic->current_language = strdup(lang_id);
	  
      if (XIM_IIIMP(im, engine_name) == NULL) {
	LangElement *langs = XIM_IIIMP(im, supported_languages);
	int count = XIM_IIIMP(im, count_languages);
	int i;
	for (i = 0; i < count; i++) {
	  if (!strcmp (langs[i].lang_id, lang_id)) {
	    /* first found entry will be default engine */
	    engine_id = langs[i].engine_id;
	    break;
	  }
	}
	if (engine_id != NULL) {
	  st = iiimcf_attr_put_string_value (attr,
					     IIIMCF_ATTR_INPUT_METHOD_NAME,
					     engine_id);
	  /* This call depends on switching policy 
	  XIM_IIIMP(im. engine_name) = engine_id;
	  */
	}
      }
    }
#endif /* HAS_IIIM_PROPERTEIS */    

    if (XIC_IIIMP(ic, context)) {
        iiimcf_destroy_context(XIC_IIIMP(ic, context));
    }       
    st = iiimcf_create_context(XIM_IIIMP(im, handle), attr, &XIC_IIIMP(ic, context));
    if (XIM_IIIMP(im, inited) == False) {
#ifdef	linux
	if (0 == strcmp(XIM_IIIMP(im, client_type), "IIIMX XIM Server")) {
	    CreateKanaWindow(im->core.display);
	}
#endif	/* linux */
	XIM_IIIMP(im, inited) = True;
    }

exit:
    iiimcf_destroy_attr(attr);
    return st;
}

void
RegisterVKBFilter(XicCommon ic) {
  _XRegisterFilterByType(ic->core.im->core.display, ic->core.focus_window,
			 ClientMessage, ClientMessage, vkb_commit_filter, (XPointer)ic);

}

void
UnregisterVKBFilter(XicCommon ic) {
  _XUnregisterFilter(ic->core.im->core.display, ic->core.focus_window,
		     vkb_commit_filter, (XPointer)ic);
}

#ifdef HAS_IIIM_PROPERTIES
Bool
IMSetPrimaryLocaleWithEngine(
    XicCommon ic,
    char *input_locale,
    char *engine_name
)
{
    char *orig;
    char *origEngine;
    XimCommon im = (XimCommon)ic->core.im;

    orig = XIM_IIIMP(im, primary_locale);
    origEngine = XIM_IIIMP(im, engine_name);
    XIM_IIIMP(im, primary_locale) = input_locale;
    XIM_IIIMP(im, engine_name) = engine_name;
    if (IMCreateIC(im, ic) != IIIMF_STATUS_SUCCESS) {
	XIM_IIIMP(im, primary_locale) = orig;
	XIM_IIIMP(im, engine_name) = origEngine;
	return False;
    }

    return True;
}
#endif /* HAS_IIIM_PROPERTIES */

Bool
IMSetPrimaryLocale(
    XicCommon ic,
    char *input_locale
)
{
    char *orig;
    XimCommon im = (XimCommon)ic->core.im;

    orig = XIM_IIIMP(im, primary_locale);
    XIM_IIIMP(im, primary_locale) = input_locale;
    if (IMCreateIC(im, ic) != IIIMF_STATUS_SUCCESS) {
	XIM_IIIMP(im, primary_locale) = orig;
	return False;
    }

    return True;
}

Bool
IMChangeClientType(
    XimCommon im,
    char *client_type
)
{
    IIIMCF_handle h;

    XIM_IIIMP(im, client_type) = client_type;
    if (0 == strcmp(XIM_IIIMP(im, client_type), "IIIMX XIM Server")) {
	xaux_proxy_setup_server(im->core.display);
    }
    if (XIM_IIIMP(im, inited)) {
	/* im has already had context,
	   so we cannot reset handle. */
	return False;
    }
    h = XIM_IIIMP(im, handle);
    if (IMCreateHandle(im) != IIIMF_STATUS_SUCCESS)
	return False;
    iiimcf_destroy_handle(h);

    return True;
}

/********************************************************************************
                       AUX load/start/done/draw/setvalues
 ********************************************************************************/

static IIIMF_status
IMSetupAuxObject(
    XimCommon im
)
{
    IIIMF_status st;
    IIIMCF_handle h;
    const IIIMCF_object_descriptor *pdesc;
    const IIIMCF_object_descriptor **ppdescs;
    IIIMCF_downloaded_object *pobjs;
    int i, n1, n2;

    h = XIM_IIIMP(im, handle);
    st = iiimcf_get_object_descriptor_list(h, &n1, &pdesc);
    if (st != IIIMF_STATUS_SUCCESS) return st;
    pobjs = (IIIMCF_downloaded_object*) malloc(sizeof(*pobjs) * n1);
    if (!pobjs) return IIIMF_STATUS_MALLOC;
    ppdescs = (const IIIMCF_object_descriptor**) malloc(sizeof(*ppdescs) * n1);
    if (!ppdescs) {
	free(pobjs);
	return IIIMF_STATUS_MALLOC;
    }

    for (i = 0, n2 = 0; i < n1; i++, pdesc++) {
	if (pdesc->predefined_id == IIIMP_IMATTRIBUTE_BINARY_GUI_OBJECT) {
	    ppdescs[n2++] = pdesc;
	}
    }
    if (n2 > 0) {
	st =  iiimcf_get_downloaded_objects(h, n2, ppdescs, pobjs);
	if (st != IIIMF_STATUS_SUCCESS) {
	    free(ppdescs);
	    free(pobjs);
	    return st;
	}
	for (i = 0; i < n2; i++) {
	    IIimpAuxDownload(pobjs[i]);
	}
    }
    free(ppdescs);
    free(pobjs);

    return IIIMF_STATUS_SUCCESS;
}

Bool
IMAuxSetValues(
    XicCommon ic,
    IIIMCF_event ev
)
{
    if (!ic) return False;

    if (iiimcf_forward_event(XIC_IIIMP(ic, context), ev)
	!= IIIMF_STATUS_SUCCESS) return False;
    IMProcessIncomingEvent(ic);

    return True;
}

Bool
IMAuxGetValues(
    XicCommon ic,
    IIIMCF_event ev
)
{
    if (!ic) return False;

    if (iiimcf_forward_event(XIC_IIIMP(ic, context), ev)
	!= IIIMF_STATUS_SUCCESS) return False;
    IMProcessIncomingEvent(ic);

    return True;
}

void
ic_change_conversion_mode (XicCommon ic, char *conv_mode)
{
  IIIMCF_event ev;

  if (ic->iiimp_icpart == NULL)
    return;

  if (conv_mode && !strcmp ((char*)conv_mode, "on"))
    {
      IIIMF_status st;
      st = iiimcf_create_trigger_notify_event (CONV_ON, &ev);
      if (st != IIIMF_STATUS_SUCCESS)
	return;
    }
  else if (conv_mode && !strcmp ((char*)conv_mode, "off"))
    {
      IIIMF_status st;
      st = iiimcf_create_trigger_notify_event (CONV_OFF, &ev);
      if (st != IIIMF_STATUS_SUCCESS)
	return;
    }
  if (iiimcf_forward_event(XIC_IIIMP(ic, context), ev) != IIIMF_STATUS_SUCCESS)
    return;
  IMProcessIncomingEvent(ic);
}

/* convert UTF-16 to UTF-8 */
char *
format_iiimcf_string (const IIIMP_card16 *u16str)
{
  iconv_t cd;
  char *inbuf, *outbuf;
  size_t inbytesleft, outbytesleft;
  IIIMP_card16 *pu;
  char *utf8str;
  size_t buflen;

  if (!u16str) return NULL;

  if ((cd = iconv_open ("UTF-8", "UTF-16")) == (iconv_t) -1)
    return NULL;

  inbytesleft = 0;
  pu = u16str;
  while (*pu) { ++pu; inbytesleft += sizeof (IIIMP_card16); }
  inbuf = (char *)u16str;

  /* The buflen should be big enough to hold the converted utf8 string */
  buflen = inbytesleft * 3;
  utf8str = (char *) malloc (buflen);
  memset (utf8str, 0, buflen);
  outbuf = utf8str;
  outbytesleft = buflen;

  iconv (cd, &inbuf, &inbytesleft, &outbuf, &outbytesleft);

  iconv_close (cd);

  return utf8str;
}

/* convert UTF-8 to UTF-16 */
static IIIMP_card16 *
format_utf8_string (char *utf8str)
{
  iconv_t cd;
  char *inbuf, *outbuf;
  size_t inbytesleft, outbytesleft;
  IIIMP_card16 *u16str;
  size_t buflen;

  if (!utf8str) return NULL;

  if ((cd = iconv_open ("UTF-16", "UTF-8")) == (iconv_t) -1)
    return NULL;

  inbuf = utf8str;
  inbytesleft = strlen (utf8str);

  /* The buflen should be big enough to hold the converted UTF-16 string */
  buflen = inbytesleft;
  u16str = (IIIMP_card16 *) malloc (sizeof (IIIMP_card16) * buflen);
  memset (u16str, 0, buflen * sizeof (IIIMP_card16));
  outbuf = u16str;
  outbytesleft = buflen * sizeof (IIIMP_card16);

  iconv (cd, &inbuf, &inbytesleft, &outbuf, &outbytesleft);

  iconv_close (cd);

  /* delete the first U+FEFF character */
#define UNICODE_BOM 0xFEFF
  if (u16str[0] == UNICODE_BOM)
    memmove (u16str, u16str + 1, (buflen - 1) * sizeof (IIIMP_card16));

  return u16str;
}

void
ic_initialize_with_input_language (XicCommon ic, char *new_lang)
{
  IIIMCF_language iiim_lang = NULL;
  IIIMF_status st;
  IIIMCF_attr attr;
  IIIMCF_event ev;
  int conversion_mode = CONV_OFF;
  char *le_name = NULL;
  int i = 0, n;
  const IIIMP_card16 *u16idname, *u16hrn, *u16domain;
  Bool found_le = False;
  IIIMCF_input_method *pims;
  XimCommon im = (XimCommon)ic->core.im;

  if (ic->iiimp_icpart == NULL)
    return;

  if (XIC_IIIMP(ic, context) == NULL)
    return;

  if (new_lang)
    {
      /* FIXME just a workaround, to be modified later */
      char *p = strchr(new_lang, ':');
      if (p) 
	{
	  *p = '\0';
	  le_name = ++p;
	}

      iiim_lang = get_input_language (ic, new_lang, True);
      if (iiim_lang != NULL)
	{
	  Xfree (ic->current_language);
	  ic->current_language = strdup (new_lang);
	}
      else
	/* invalid new lang */
	return;
    }
  else
    iiim_lang = get_input_language (ic, ic->current_language, False);

  if (le_name != NULL)
    {
      char *idname = NULL;
      /* probably le_name consists of 'lename' and 'imename' which are 
       * separated by '-'.
       */
      const char *ime_delimiter = "-";
      extern char **my_strsplit(const char *, const char*);
      char **le_ime = my_strsplit (le_name, ime_delimiter);
      char *lename = le_ime[0], *imename = le_ime[1];

#ifdef HAS_IIIM_PROPERTIES		  
      free (ic->current_le);
      ic->current_le = strdup (le_name);
#endif /* HAS_IIIM_PROPERTIES */		  

      st = iiimcf_get_supported_input_methods (XIM_IIIMP (im, handle), &n, &pims);
      if (st != IIIMF_STATUS_SUCCESS)
	return;

      for (i = 0; i < n; i++)
	{
	  st = iiimcf_get_input_method_desc (pims[i], &u16idname, &u16hrn,
					     &u16domain);
	  if (st != IIIMF_STATUS_SUCCESS)
	    return;

#ifdef HAS_IIIM_PROPERTIES
	  idname = format_iiimcf_string (u16idname);
#else
	  idname = format_iiimcf_string (u16hrn);
#endif /* HAS_IIIM_PROPERTEIS */	  
	  
	  if (idname != NULL && strcmp (lename, idname) == 0)
	    {
	      /* update the pims[i]->imname with the new imname that
	       * consists imename.
	       */
	      if (imename)
		{
		  IIIMP_card16 *imname = format_utf8_string (le_name);
		  iiimcf_set_input_method_name (pims[i], imname);
		}

	      free (idname);
	      found_le = True;
	      break;
	    }
	  if (idname != NULL)
	    {
	      free (idname);
	      idname = NULL;
	    }
	}
      if (lename) free (lename);
      if (le_ime) free (le_ime);
    }

  /* save conversion mode to restore this after switching */
  st = iiimcf_get_current_conversion_mode (XIC_IIIMP(ic, context),
					   &conversion_mode);

  st = iiimcf_create_attr (&attr);
  if (st != IIIMF_STATUS_SUCCESS) 
    return;

  if (found_le)
      iiimcf_attr_put_ptr_value (attr, IIIMCF_ATTR_INPUT_METHOD,
				 pims[i]);
  if (iiim_lang)
      iiimcf_attr_put_ptr_value (attr, IIIMCF_ATTR_INPUT_LANGUAGE,
				 iiim_lang);
#ifdef HAS_IIIM_PROPERTIES
  iiimcf_attr_put_integer_value (attr, IIIMCF_ATTR_KBD_LAYOUT,
				 GetKBDLayout ());
#endif /* HAS_IIIM_PROPERTIES */

  st = iiimcf_context_set_attr(XIC_IIIMP(ic, context), attr);
  iiimcf_destroy_attr (attr);
  if (st != IIIMF_STATUS_SUCCESS)
    return; 

  /* NOTE: need to support the change of input language here.
   * and don't call ic_switcher_set_input_language here too.
   * it causes the infinite loop of PropertyNotify event
   * between xiiimp.so and gimlet.
   */
  /* ic_switcher_set_input_language(ic, NULL); */
  st = iiimcf_create_seticfocus_event (&ev);

  if (st != IIIMF_STATUS_SUCCESS)
    return;

  iiimcf_forward_event(XIC_IIIMP(ic, context), ev);
  if (conversion_mode)
    {
      st = iiimcf_create_trigger_notify_event (CONV_ON, &ev);
      if (st != IIIMF_STATUS_SUCCESS)
	return;
      iiimcf_forward_event(XIC_IIIMP(ic, context), ev);
    }
  IMProcessIncomingEvent(ic);
}

/* get Lang List */
IIIMCF_language*
iiim_get_languages (XimCommon im, int *n_lang)
{
  IIIMF_status st;
  IIIMCF_handle h = XIM_IIIMP(im, handle);
  IIIMCF_language *lang = NULL;

  if (!h)
    return NULL;

  st = iiimcf_get_supported_languages (h, n_lang, &lang);
  if (st != IIIMF_STATUS_SUCCESS)
    return NULL;

  return lang;
}

static IIIMCF_language
get_input_language (XicCommon ic, char *input_language, Bool exact_match)
{
  int i;
  IIIMF_status st;
  char *langid, *lang, *p;
  int n_lang;
  IIIMCF_language *lang_list;
  XimCommon im = (XimCommon)ic->core.im;

  if (input_language == NULL)
      return NULL;
  else
      lang = strdup(input_language);

  lang_list = iiim_get_languages (im, &n_lang);
  if (lang_list)
    {
      for (i = 0; i < n_lang; i++)
	{
	  st = iiimcf_get_language_id(lang_list[i], (const char **)&langid);
	  if (st != IIIMF_STATUS_SUCCESS)
	    continue;
	  if (!strcmp(lang, langid))
	    goto found_language;
	}
      /* try to find the supported language again without charset locale */
      p = strchr(lang, '@');
      if (p)
	*p = 0;
      p = strchr(lang, '.');
      if (p)
	*p = 0;

      for (i = 0; i < n_lang; i++)
	{
	  st = iiimcf_get_language_id(lang_list[i], (const char **)&langid);
	  if (st != IIIMF_STATUS_SUCCESS)
	    continue;
	  if (!strcmp(lang, langid))
	    goto found_language;
	}
      /* finally try to take care of the locale, such as en */
      p = strchr(lang, '_');
      if (p)
	*p = 0;

      for (i = 0; i < n_lang; i++)
	{
	  st = iiimcf_get_language_id(lang_list[i], (const char **)&langid);
	  if (st != IIIMF_STATUS_SUCCESS)
	    continue;
	  if (!strcmp(lang, langid))
	    goto found_language;
	}

      /*
	still cannot find a match. Just take one for "en" as
	a default.
      */
      for (i = 0; i < n_lang; i++)
	{
	  st = iiimcf_get_language_id (lang_list[i],
				       (const char **)&langid);
	  if (st != IIIMF_STATUS_SUCCESS)
	    continue;
	  if (!strncmp (langid, "en", 2))
	    goto found_language;
	}
    }

  free(lang);
  return NULL;

 found_language:;
  free(lang);
  Xfree(ic->current_language);
  ic->current_language = strdup(langid);

  return lang_list[i];
}

/* Local Variables: */
/* c-file-style: "iiim-project" */
/* End: */
